/************************************************************************/
/*                                                                      */
/*   Program Name :  k5tovdif                                           */
/*                                                                      */
/*     K5tovdif is a K5/VSSP32 to VDIF format converter                 */
/*                                                                      */
/*     Enviroment variables                                             */
/*           VDIFDIR   ---   default directory for VDIF data            */
/*                                                                      */
/*            Ver 1.01  2016-03-08                                      */
/*                                                                      */
/*      Copyright (c) 2014-2016 T.Kondo All Right Reserved              */
/*                                                                      */
/************************************************************************/
/*
     2014-06-10  -- development started 
     2014-06-13  -- Ver. 1.00 
     2016-03-08  -- Ver. 1.00 


   how to run
          k5tovdif k5name1 [k5name2 k5name3 k5name4 [vdifname]] [options]
          

    
    where    
             
             k5name1 -- K5 data file name containing Ch01-04 or Ch01
             k5name2 -- K5 data file name containing Ch05-08 or Ch02
             k5name3 -- K5 data file name containing Ch09-12 or Ch03
             k5name4 -- K5 data file name containing Ch13-16 or Ch04
                  2nd and after K5 file names can not be accompanied by 
                  directory name, the directory of 1st K5 file is 
                  adopted for these files

             vdifname  -- VDIF file name to be created

             options (any order)
               -o vdifname     -- VDIF file name to be created
                                 (default : extension of 1 st K5 name is replaced by ".vdif")
               -dir vdifdir    -- directory for VDIF file out
                                 (default  : K5 directory)
               -so strat_sec  -- offset time from data head (sec)
                                 (default :  0 )
               -p period     -- data period to convert (sec)
                                 (default : all)
               -das das_name --- set station name or compulsoly set station name
                                 (in case of VSSP32 station or PC name got from it's AUX data)
               -sid sid      --- set station id (A2)

               -date jjj|yyyymmdd|yyyyddd  --   set date in case of VSSP32
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year

               -e2bit n    -- 2bit data encode mode
                               n=1 : offset binary 0 1 2 3
                               n=2 : twisted offset binary 0 2 1 3
                               n=3 : signed binary 1 0 2 3
                                     (default is 2)
               -fr[amelength] frame_bytes -- set frame length without header (in byte)
	                                         (default 1280)
               -num[frame_sec] nframe -- Set # of frames per second
                                (default: 2048e6/(1280*8) = 200000)
               -nocheck      -- no header error check (adopted for 2nd and later headers)

                **** follwoing will be implemented *****

               -ch1 nn       -- set K5 channel corresponding to VDIF ch #1
                                     nn is K5 channel number (1-16)
               -ch2 nn       -- set K5 channel corresponding to VDIF ch #2
                   ...........................
               -ch16 nn      -- set K5 channel corresponding to VDIF ch #16

               -chall n1:n2:....:n16    -- set K5 channels corresponding to VDIF all channels
                               n1  -- K5 channel number (1-16) corresponding to VDIF ch 1
                               n2  -- K5 channel number (1-16) corresponding to VDIF ch 2
                                 ...............
                               n16 -- K5 channel number (1-16) corresponding to VDIF ch 16

  Enviroment variables
          VDIFDIR  ---   default path to VDIF data


*** sample run parameters ****
*/

#define PROGRAM_VERSION "k5tovdif (Ver 1.01  2016-03-08)"


//#define ORIGINAL    // for debug

//#include "libk5.h"      // common header for IP-VLBI utilities
//#include "libmk5tok5.h"  // library for Mark5 to K5 conversion
//#include "libvssp32.h"  // library for Mark5 to K5 conversion
#include "libvdif.h"

/* default conversion information file name set */
#define VDIFDIR "VDIFDIR" // the environment variable for VDIF file

/* default parameters  */
#define K5NUMCH  4
#define STATION_NAME "kashima" // station name at most 8 characters (A8)
#define EDV_NUMBER  1     // Extended format version number "1" for Kashima VDIF
#define FRAME_BYTES 1280  // Frame lenghth without header (in bytes)

#define DFLT_DATA_FILE "vdif.dat"
//#define BUFMAX 25000   // maximum read buffer size
#define FRAME_LENGTH_MAX 1500   // maximum output buffer size (usually packet length) (bytes)


typedef struct k5tovdifrunp     // run parameters for k5 to VDIF conversion program
{
    int ptype;   // run parameter type  =1  mode1,  =2 mode 2
	int headlen;
    int channel;
    int group;
    int soffset;
    int period;
    int reverse;
    int makemode;
    int scan;    // scan# for get MODE from VEX file (default=1) 
    int odd;    // =1 4byte integer offset mode in case of 64 track
    int monit;   // =1  inofo_file monitor ON,  =0 no monitoring
    int yyyy, month, dd, ddd;  // observation date year month,day of month, day of year
	int hh,mm,ss; // UTC
    int mjd;   //  MJD
    int mjd3;   // lower 3 digit of MJD
    int k5num;
    char vdifname[256];
    char k5name[4][256];
    char outdir[256];
    char info_file[256];
    char vex_file[256];
    char sid[3];
	char das[9];
    char program_version[60];
	int frame_bytes; // # of bytes in a frame (excepting header)
	int nframe_sec;  // # of frames per second
    // for mode2 only run parameres are as follows
    int ibit[64];
    int nbits;
	int e2bit;
    int nocheck;
    double fmhz;
	int ch_assign[16];  // channel assign table 

    // internal use parameers 
	int imax;
	int nbuf;   // obuf size
    int totch;  // total number of channels
    int bitsample;  // # of bits/sample
    double sec_period[4];  // contain actual data size for 4 files (in sec)
    int isec_period[4];
    int K5file_id[16];  // to show file id (by 1-4) correspond to K5 ch# 1-16
    int K5chan_id[16];  // to show chanel # in K5 file (by 1-4) correspond to K5 ch# 1-16 
	int sfreq;
    int numch;
    int adbit;
    double fsampl;
	double tsampl;
	double unitp;
    int usampl;
    int bits, bytes, numb;
	int samplpp;
    int k5nbuf;  // read size at K5 file
    int adjust_flag;   // flag to set conversion length integer second  1:set 0: no set

} k5tovdifrunp_t;


void replace_extension(char *fname, char *extension)
{
/*
     replace file name extension by given extension
	 if extension portion is existed in the original name.
	 if it is not existed, extension is added as an extension
*/
	int i,n,len,len2;
	len=strlen(fname);
	len2=strlen(extension);
	n=len-1;
	while(n >=0){
		if(fname[n]=='.'){
			for(i=0; i<len2; i++){
				fname[n+i+1]=extension[i];
			}
			fname[n+len2+1]='\0';
			goto exit2return;
		}
		n--;
	}
	fname[len]='.'; fname[len+1]='\0';
	strcat(fname,extension);
exit2return:;
}


#ifndef _LIBK5_

int hhmmss2sec(char *cbuf, int *hh, int *mm, int *ss)
{
    //  HH:MM:SS => seconds
    int n,len;
    int hh1,mm1,ss1,seconds;
    int p1,p2, n1st;
    char buf[10];
    len=strlen(cbuf);
    // check ':' position
    p1=0; p2=0;
    n1st=0;
    //printf("len %d\n",len);
    for(n=0; n<len; n++){
        if(cbuf[n]==':'){
            if(n1st==0){
                p1=n;
                n1st=1;
            } else {
                p2=n;
                n1st=2;
            }
        }
    }
    //printf("p1,p2 %d %d\n",p1,p2);
    if(n1st==0){ // no HH:MM:SS
       seconds=atoi(cbuf);
       hh1=seconds/3600;
       mm1= (seconds % 3600)/60;
       ss1= seconds % 60;
       *hh=hh1; *mm=mm1; *ss=ss1;
       return seconds;
    }

    strncpy(buf,cbuf,p1);
    buf[p1]='\0';
    hh1=atoi(buf);
    strncpy(buf,cbuf+p1+1,p2-p1-1);
    buf[p2-p1]='\0';
    mm1=atoi(buf);
    strncpy(buf,cbuf+p2+1,len-p2-1);
    buf[len-p2]='\0';
    ss1=atoi(buf);
    seconds=hh1*3600+mm1*60+ss1;
    *hh=hh1; *mm=mm1; *ss=ss1;
    //printf("HH MM SS %d %d %d\n",hh,mm,ss); // debug
    return seconds;
}



int leaf_year_check(int year)   // return 0 in case of leaf year
{
       /* leaf year check */
      if (year-(year/4)*4 == 0) 
      { 
          if (year-(year/100)*100 == 0) 
          { 
               if (year-(year/400)*400 == 0) { return 0;}
               return -1;
           }
           return 0;
       }
       return -1;
}

void ddd2md(int year, int ddd, int* month, int* mday)
{
      int md[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
      int i,td,wd ;
       /* leaf year check */
      if (leaf_year_check(year)==0)
      {
         md[1]=29;
      }
      *month=0;
      *mday=0;
      td=ddd;
      for (i=0; i< 12; i++)
      {
         wd=td-md[i];
         if (wd <=0)
         {
           *mday=td;
           *month=i+1;
           goto exithere;
          }
         td=wd;
       }
   exithere:;
 }

void md2ddd(int year, int* ddd, int month, int mday)
{
      int md[12] = {0,31,28,31,30,31,30,31,31,30,31,30};
      int i;
       /* leaf year check */
      if (leaf_year_check(year)==0)
      {
         md[2]=29;
      }
      *ddd=0;
      for (i=0; i< month; i++)
      {
        *ddd=*ddd+md[i];
      }
      *ddd += mday;
 }

/* UTC to MJD conversion */
int ut2mjd(double *d, int year, int month, int day,
           int hour, int min, int sec)
/*
  [in]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
  [return]
     ut2mjs -- MJD  (integer day)
     d      -- fractional part of MJD (day)
*/
{
    int  bc, gregory, mnth, y4; 
    //long j;
    int j;

	mnth=month;
	y4=year;
    bc = year <= 0;
    
    /* First day at Gregory's calendar */
    gregory = year > 1582 || (year == 1582 && month > 10) ||
              (year == 1582 && month == 10 && day >= 15);
            
    if (mnth <= 2) {
        y4--;
        mnth += 12;
    }
    
    if (hour < 12) {
        j = 0;
        *d = 0.5;
    } else {
        j = 1;
        *d = -0.5;
    }
    *d += (hour*3600L + min*60 + sec)/86400.0;
    
    j += bc ? (y4-3)/4 : y4/4;
    if (gregory) j+= 2 - y4/100 + y4/400;
    j += 1720994L + y4*365L + (mnth+1)*30 + (mnth+1)*3/5 + day;
    
	/* convert to MJD */
	if (hour < 12) {
		*d -= 0.5;
	} else {
		*d += 0.5;
		j -= 1L;
	}
    j -= 2400000L;
    return j;
}

/* MJD to UTC conversion */
void mjd2ut(int *year, int *month, int *day,
		   int *hour, int *min, int *sec, int j, double d)
/*
  [in]
     j      -- MJD  (integer day)
     d      -- fractional part of MJD (day)
  [return]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
*/
{
    int c, k, e, s;

    /* convert to JD */
    if( d >= 0.5) {
		d -= 0.5;
		j++;
	} else {
		d += 0.5;
	}
	j += 2400000L;

	/* JD to UTC */
    if (d >= 0.5) {
        j++;
        d -= 0.5;
    } else d += 0.5;

    if (j >= 2299161) j = j + 1 + (j-1867216)/36524 - (j-1867216)/146096;
    j += 1524;

    //c = (long)((j - 122.1)/365.25);
    c = (int)((j - 122.1)/365.25);
    k = 365L*c + c/4;
    //e = (long)((j - k)/30.6001);
    e = (int)((j - k)/30.6001);

    *year = c - 4716;
    *month = e - 1;
    if (*month > 12) {
       (*month) -= 12;
       (*year)++;
    }
    //*day = j - k - (long)(30.6*e);
    *day = j - k - (int)(30.6*e);

    //s = (long)(d * 86400 + 0.5);
    s = (int)(d * 86400 + 0.5);
    *hour = s / 3600;
    *min = (s % 3600) / 60;
    *sec = s % 60;
}

#endif


int vex_parse(char *cbuf, vparam_t *vparam)  //2003-03-21
{
/*
    Parse VEX parameters by '=',':',';', and ' '.
    '=',':',';' are returned as a parsed item for further analysis
    Total number of items including '=',':',';' is defined in
    type define statesment of vparam as

    typedef struct vparam
    {
        int  n;          //number of param
        char param[60][128];   // 60x128 characters
    } vparam_t;

    In this case 60 is a maximum number of items and
    128 characters are maximum char in an item.

    If runtime error occurrs, please try to increase the number of
    items in the vparam definition.


*/

    // vex_parse =0;  single statement in a cbuf
    //           =1;  more statement left in a cbuf, start point in cbuf 
    //                is updated 
    char abuf[500];
    char *cset=" ";
    char *p=NULL;
    int i,ipos,k,kpos,flag,mflag,len;

    len=strlen(cbuf);
    //printf("len = %d\n",len);

    //remove extra space in front

    if(cbuf[0]==' '){
        kpos=0;
        for(k=0; k<len; k++){
            if(cbuf[k]!=' '){
                kpos=k;
                break;
            }
        }
        i=0;
        for(k=kpos; k<len; k++){
            cbuf[i]=cbuf[k];
            i++;
        }
        cbuf[i]='\0';
    }
    // len again

    len=strlen(cbuf);


    // parse start

    if(cbuf[0] == '*') { // comment
        vparam->n=0; 
        return 0;
    }
    i=0;
    ipos=0;
    flag=0;
    for(k=0; k<len; k++){
        if(cbuf[k] == ' ' || cbuf[k] == '=' || cbuf[k] == ':'){
            if(flag==1) {  // one parameter end
                vparam->param[ipos][i]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            if(cbuf[k] != ' '){
                vparam->param[ipos][i]=cbuf[k];
                vparam->param[ipos][i+1]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            goto nextk;
        }
        if(cbuf[k] == ';'){
            kpos=k;
            goto exithere;
        }
        if(cbuf[k] != ' '){
            flag=1;
            vparam->param[ipos][i]=cbuf[k];
            i++;
        }
nextk:;
    }
    kpos=k;
exithere:
    vparam->param[ipos][i]='\0';
    ipos++;
    vparam->n=ipos;
    i=0;
    mflag=0;
    if(kpos < len-1){
        for(k=kpos+1; k<len; k++) {
            if(cbuf[k] != '\0'){
                abuf[i]=cbuf[k];
                i++;
            }
            if(cbuf[k] == ';') mflag=1;
        }
    }
    if(mflag==1){
        abuf[i]='\0';
        strcpy(cbuf,abuf);
        return 1;
    }
    return 0;
}



void param_get_k5tovdif(int argc, char *argv[], k5tovdifrunp_t *runp)                       
{
    int n,i;
    int len;
    int mode;
    char cbuf[256],buf[10];
    double dw; // work
    vparam_t vp;

    // initialize parameters

    /* common parameters */
    runp->ptype=1;  // run with information file
    for(n=0; n<4; n++) runp->k5name[n][0]='\0';
    runp->vdifname[0]='\0';

    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->reverse=0;
    runp->makemode=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->das[0]='\0';
    runp->scan=1;
    runp->monit=0;

    /* mode 2 parameters */
    runp->numch=K5NUMCH;
    runp->e2bit=2;
    for(n=0; n<32; n++) runp->ibit[n]=0;

    runp->yyyy=0; runp->month=0; runp->dd=0; runp->ddd=0;
    runp->mjd=0; runp->mjd3=0;

    runp->nbits=32;
	runp->frame_bytes=FRAME_BYTES;
    runp->nframe_sec=0;
	runp->nframe_sec=0;
	runp->headlen=8;  // fixed value

    runp->yyyy=0;
    runp->ddd=0;
    runp->month=0;
    runp->dd=0;
    runp->hh=0;
    runp->mm=0;
    runp->ss=0;

    runp->nocheck=0; 
	for(n=0; n<16; n++) runp->ch_assign[n]=n+1; 

    for(i=0; i<9; i++) runp->das[i]='\0';
    for(i=0; i<3; i++) runp->sid[i]='\0';

//	if(strlen(STATION_NAME) > 8){
//		strncpy(runp->das,STATION_NAME,8);
//	} else {
//		strcpy(runp->das,STATION_NAME);
//	}

    mode=0;
    /* 1st param check whether it is started with "-" or not  */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        if(cbuf[0] == '-') mode=1;
    }

    if(mode == 1) {
        n=2;
        goto comm_anal;
    }


    n=2;
    runp->k5num=0;

    /* K5 name get */
    while(1){
        if(argc >= n) {
           strcpy(cbuf,*(argv+ n-1) );
           //printf("%s\n",cbuf);
           if(cbuf[0] == '-') break;  // go to command parameter get
           //if(runp->k5num >= 4) {
           //    printf(" Warning!! Number of K5 file exceeds 4!!\n");
           //    printf(" So more K5 files will be neglected!!\n");
           //    runp->k5num=4;
           //    n++;
           //    break;
           //}
           if(runp->k5num >= 4) {  // mark5 name
               strcpy(runp->vdifname,cbuf);
               runp->k5num=4;
               n++;
               break;
           }
           strcpy(runp->k5name[runp->k5num],cbuf);
           runp->k5num++;
        } else {
            goto exitend;
        }
        n++;
    }


comm_anal:

	if(cmnd_get_n(argc,argv,"-o",cbuf)==0){     // VDIF name
		strcpy(runp->vdifname,cbuf);
        /* check Mark5 file name is acompanied by directory or not */
        get_true_path(runp->vdifname, VDIFDIR, "./",cbuf);
        strcpy(runp->vdifname,cbuf);
	}
	//if(cmnd_get_n(argc,argv,"-d",cbuf)==0){     // out directory
	if(cmnd_get_n(argc,argv,"-dir",cbuf)==0){     // out directory   2016-03-08
        strcpy(runp->outdir,cbuf);
        /* env variable is substituted by real one */
        if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
            strcpy(runp->outdir,cbuf);
        }
    }
    if(cmnd_get_n(argc,argv,"-e",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2 bit encode mode
    //if(cmnd_get_p(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf); // start offset
    if(cmnd_get_p(argc,argv,"-so",cbuf)==0) runp->soffset = atoi(cbuf); // start offset  2016-03-08
    if(cmnd_get_n(argc,argv,"-sid",cbuf)==0) {                         // station id
        runp->sid[0]=cbuf[0]; runp->sid[1]=cbuf[1];
    }
    if(cmnd_get_n(argc,argv,"-p",cbuf)==0) runp->period = atoi(cbuf);  // period
    if(cmnd_get_n(argc,argv,"-nocheck",cbuf)==0) runp->nocheck=1;      // no check mode
    if(cmnd_get_n(argc,argv,"-mon",cbuf)==0) runp->monit=1;            // monitor mode on
    if(cmnd_get_p(argc,argv,"-ch1",cbuf)==0) runp->ch_assign[0]=atoi(cbuf);  // CH1 set
    if(cmnd_get_n(argc,argv,"-ch01",cbuf)==0) runp->ch_assign[0]=atoi(cbuf);  // CH1 set
    if(cmnd_get_n(argc,argv,"-ch2",cbuf)==0) runp->ch_assign[1]=atoi(cbuf);  // CH2 set
    if(cmnd_get_n(argc,argv,"-ch02",cbuf)==0) runp->ch_assign[1]=atoi(cbuf);  // CH2 set
    if(cmnd_get_n(argc,argv,"-ch3",cbuf)==0) runp->ch_assign[2]=atoi(cbuf);  // CH3 set
    if(cmnd_get_n(argc,argv,"-ch03",cbuf)==0) runp->ch_assign[2]=atoi(cbuf);  // CH3 set
    if(cmnd_get_n(argc,argv,"-ch4",cbuf)==0) runp->ch_assign[3]=atoi(cbuf);  // CH4 set
    if(cmnd_get_n(argc,argv,"-ch04",cbuf)==0) runp->ch_assign[3]=atoi(cbuf);  // CH4 set
    if(cmnd_get_n(argc,argv,"-ch5",cbuf)==0) runp->ch_assign[4]=atoi(cbuf);  // CH5 set
    if(cmnd_get_n(argc,argv,"-ch05",cbuf)==0) runp->ch_assign[4]=atoi(cbuf);  // CH5 set
    if(cmnd_get_n(argc,argv,"-ch6",cbuf)==0) runp->ch_assign[5]=atoi(cbuf);  // CH6 set
    if(cmnd_get_n(argc,argv,"-ch06",cbuf)==0) runp->ch_assign[5]=atoi(cbuf);  // CH6 set
    if(cmnd_get_n(argc,argv,"-ch7",cbuf)==0) runp->ch_assign[6]=atoi(cbuf);  // CH7 set
    if(cmnd_get_n(argc,argv,"-ch07",cbuf)==0) runp->ch_assign[6]=atoi(cbuf);  // CH7 set
    if(cmnd_get_n(argc,argv,"-ch8",cbuf)==0) runp->ch_assign[7]=atoi(cbuf);  // CH8 set
    if(cmnd_get_n(argc,argv,"-ch08",cbuf)==0) runp->ch_assign[7]=atoi(cbuf);  // CH8 set
    if(cmnd_get_n(argc,argv,"-ch9",cbuf)==0) runp->ch_assign[8]=atoi(cbuf);  // CH9 set
    if(cmnd_get_n(argc,argv,"-ch09",cbuf)==0) runp->ch_assign[8]=atoi(cbuf);  // CH9 set
    if(cmnd_get_n(argc,argv,"-ch10",cbuf)==0) runp->ch_assign[9]=atoi(cbuf);  // CH10 set
    if(cmnd_get_n(argc,argv,"-ch11",cbuf)==0) runp->ch_assign[10]=atoi(cbuf);  // CH11 set
    if(cmnd_get_n(argc,argv,"-ch12",cbuf)==0) runp->ch_assign[11]=atoi(cbuf);  // CH12 set
    if(cmnd_get_n(argc,argv,"-ch13",cbuf)==0) runp->ch_assign[12]=atoi(cbuf);  // CH13 set
    if(cmnd_get_n(argc,argv,"-ch14",cbuf)==0) runp->ch_assign[13]=atoi(cbuf);  // CH14 set
    if(cmnd_get_n(argc,argv,"-ch15",cbuf)==0) runp->ch_assign[14]=atoi(cbuf);  // CH15 set
    if(cmnd_get_n(argc,argv,"-ch16",cbuf)==0) runp->ch_assign[15]=atoi(cbuf);  // CH16 set

    if(cmnd_get_n(argc,argv,"-chall",cbuf)==0) { // all channel set
        vex_parse(cbuf, &vp);
        i=0;
        for(n=0; n<vp.n; n++) {
           //printf(" n= %d  %s\n",n,vp.param[n]);// debug
            if(vp.param[n][0]!=':'){
                runp->ch_assign[i]=atoi(vp.param[n]);
                i++;
            }
        }

    }
    if(cmnd_get_n(argc,argv,"-das",cbuf)==0) { // station name set
        if(strlen(cbuf)<9){
            strncpy(runp->das,cbuf,strlen(cbuf));
        } else {
            strncpy(runp->das,cbuf,8);
        }
    }


    if(cmnd_get_n(argc,argv,"-date",cbuf)==0) { // date set
        len=strlen(cbuf);
//        //printf("len = %d\n",len);
        if(len==8){ // yyyymmdd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,2); buf[2]='\0';
            runp->mm=atoi(buf);
            strncpy(buf,cbuf+6,2); buf[2]='\0';
            runp->dd=atoi(buf);
            //get MJD
            runp->mjd=ut2mjd(&dw,runp->yyyy,runp->month,runp->dd,0,0,0);
            //get total day 
            md2ddd(runp->yyyy,&runp->ddd,runp->month,runp->dd);
        } else if(len==7){ // yyyyddd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,3); buf[3]='\0';
            runp->ddd=atoi(buf);
            //get month day
            ddd2md(runp->yyyy,runp->ddd,&runp->month,&runp->dd);
            //get MJD
            runp->mjd=ut2mjd(&dw,runp->yyyy,runp->month,runp->dd,0,0,0);
        } else {
            runp->mjd = atoi(cbuf);
        }
        runp->mjd3 = runp->mjd % 1000;
    }


exitend:;

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("How to use\n");
    printf("\n");
    //printf("    k5tom5 yymmdd k51name [k52name k53name k54name] [options]\n");
    printf("    k5tovdif k5name1 [k5name2 k5name3 k5name4 [vdifname]] [options]\n");
    printf("\n");
    printf("   where\n");
    printf("       k5name1 ---- K5 file name for CH01-04 or CH01\n");
    printf("       k5name2 ---- K5 file name for CH05-08 or CH02\n");
    printf("       k5name3 ---- K5 file name for CH09-12 or CH03\n");
    printf("       k5name4 ---- K5 file name for CH13-16 or CH04\n");
    printf("                      2nd and after K5 file names are not accompanied by\n");
    printf("                      directory name, the directory of 1st K5 file is \n");
    printf("                      adopted for these files\n\n");
    printf("       vdifname  ---- VDIF file name to be created\n\n");
    printf("      options (any order)\n");
    printf("        -o vdifname     -- VDIF file name to be created\n");
    printf("             (default : extension of 1st K5 name is replaced by \'.vdif\')\n");
    printf("        -dir vdifdir   -- directory for VDIF file out\n");
    printf("                           (default  : K5 directory)\n");
    printf("        -so soffset     -- offset time from data head (sec)\n");
    printf("                          (default :  0 )\n");
    printf("        -p period     -- data period to convert (sec)\n");
    printf("                          (default : all)\n");
    printf("        -das das_name --- set station name or compulsoly set station name\n");
    printf("                 (in case of VSSP32 station or PC name got from it\'s AUX data)\n");
    printf("        -sid sid      --- set station id (A2)\n");
    printf("                 (in case of VSSP32 sid got from it\'s AUX data)\n");
    printf("        -date jjj|yyyymmdd|yyyyddd  --   set date in case of VSSP\n");
    printf("                     jjj      : lower 3digits of MJD \n");
    printf("                       or\n");
    printf("                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month\n");
    printf("                       or\n");
    printf("                     yyyyddd  : 4 digit year, 3 digit day of year\n");
    printf("        -e2bit n    -- 2bit data encode mode\n");
    printf("                        n=1 : offset binary 0 1 2 3\n");
    printf("                        n=2 : twisted offset binary 0 2 1 3\n");
    printf("                        n=3 : signed binary 1 0 2 3\n");
    printf("                          (default is 2)\n");
    printf("        -nocheck    -- no header error check\n"); 
    printf("                          (adopted for 2nd and later headers)\n"); 
    printf("        -fr[amelength] frame_bytes -- set frame length without header (in byte)\n");
    printf("	                       (default 1280)\n");
    printf("        -num[frame_sec] nframe -- Set # of frames per second\n");
    printf("                                 (default: 2048e6/(1280*8) = 200000)\n");
    printf("                **** follwoing will be implemented *****\n");
    printf("        -ch1 nn     -- set K5 channel corresponding to Mark5 ch #1\n"); 
    printf("                          nn is K5 channel number (1-16)\n"); 
    printf("        -ch2 nn     -- set K5 channel corresponding to Mark5 ch #2\n"); 
    printf("           ...........................\n"); 
    printf("        -ch16 nn    -- set K5 channel corresponding to Mark5 ch #16\n"); 
    printf("        -chall n1:n2:....:n16\n");
    printf("                 -- set K5 channels corresponding to Mark5 all channels\n"); 
    printf("                     n1  -- K5 channel number (1-16) corresponding to Mark5 ch 1\n"); 
    printf("                     n2  -- K5 channel number (1-16) corresponding to Mark5 ch 2\n"); 
    printf("                      ...............\n"); 
    printf("                     n16 -- K5 channel number (1-16) corresponding to Mark5 ch 16\n"); 
    //printf("        -monit        -- info_file monitor ON\n");
    printf("\n");
}

void monit_runp(k5tovdifrunp_t *runp, vdif_header_t *vhead)
// run parameter monitor
{
    int n;

       printf(" ---------- RUN CONDITION ----------\n");
       printf("   yyyymmdd  : %04d%02d%02d\n",runp->yyyy,runp->month,runp->dd);
       printf("   yyyyddd   : %04d%03d\n",runp->yyyy,runp->ddd);
       printf("   jjj       : %03d\n",runp->mjd3);
       if(runp->yyyy != 0){
       printf("   MJD       : %d\n",runp->mjd);
       }
       for(n=0; n<runp->k5num; n++)
       printf("   k5name %d : %s\n",n+1,runp->k5name[n]);

       printf("   vdifname  : %s\n",runp->vdifname);
       printf("   outdir    : %s\n",runp->outdir);
       //printf("   infofile  : %s\n",runp->info_file);
       //printf("   channel   : %d\n",runp->channel);
       //printf("   group     : %d\n",runp->group);
       printf("   soffset   : %d\n",runp->soffset);
       printf("   period    : %d\n",runp->period);
       printf("   e2bit     : %d\n",runp->e2bit);
       printf("   nocheck   : %d\n",runp->nocheck);
       printf("   VDIF ch   : %d\n",runp->totch);
//       printf("     ----- channel assign -----\n");
//       printf("     M5b CH#  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16\n");
//       printf("     K5  CH#");
//       for(n=0; n<16; n++) printf("%3d",runp->ch_assign[n]);
//       printf("\n");
	   printf("   station id (A2) : %s\n",runp->sid);
	   printf("   station das (A8): %s\n",runp->das);
	   printf("   Frame length (without header) : %d bytes\n",runp->frame_bytes);
	   printf("   Frame length (with header)    : %d bytes\n",runp->frame_bytes+runp->headlen*4);
	   printf("   # of frames per second        : %d\n",runp->nframe_sec);
	   printf("   start time : %4d/%02d/%02d %02d:%02d:%02d\n",
		   vhead->t.year,vhead->t.month,vhead->t.mday,vhead->t.hh,vhead->t.mm,vhead->t.ss);
	   printf("   start ref epo:  %d,   sec from epoch: %d\n", vhead->ref_epo,vhead->sec_fm_ref);
       //printf("   odd       : %d\n",runp->odd);
       //printf("   parity    : %d\n",runp->parity);
       //printf("   nrzm      : %d\n",runp->nrzm);
       //printf("   vlba      : %d\n",runp->vlba);
       //printf("   nbits     : %d\n",runp->nbits);
       //if(runp->vex_file[0] != '\0')
       //   printf("   vex_file  : %s\n",runp->vex_file);
       //if(runp->sid[0] != '\0' && runp->sid[0] !='-')
       //   printf("   station_id: %s\n",runp->sid);
       //printf("   modulation: %d\n",runp->modulation);
       printf(" -------------------------------------\n");

    printf("\n");
}

void setup_k5files_full_path(k5tovdifrunp_t *runp)
{    // 2008-09-30
/*
   full path of K5 filename is setupped using the directory
   of 1 st K5 name if 2nd and after K5 files are not acompanied 
   by directory

*/
    char dir0[128],dir[128],fname[128];
    int n;
    parse_dir_and_name(runp->k5name[0], dir0, fname);
    if(dir0[0]!='\0'){
        for(n=1; n<runp->k5num; n++){
            parse_dir_and_name(runp->k5name[n], dir, fname);
            if(dir[0]=='\0'){
               strcpy(runp->k5name[n],dir0);
               strcat(runp->k5name[n],fname);
            }
        }
    }
}

void parseADdataKOKO(u_int32_t *ia, int numb, int adbit, int *xdata, int *npos)
{
/*
    IP-VLBI{[hf[^A/D\ɉTvOlɕϊ
      ch͂̎_ł͊֌WȂ

      [in]
         ia  ---  sampler (IP-VLBI board) out data 
         numb --  # of elements of ia 
         adbit -- sampler A/D resolution bits (1,2,4,8)
      [return]
         xdata -- data parsed into each sampled data
         npos  -- valid number of data in xdata array

     Relation between # of ch (numch) and each ch time siries

        chnumchƂ
        ch# mk Ԗڂ̃Tvf[^

        sdata(m,k)=xdata(k*numch+m)

        ƂȂ
        
*/
     int numb32, pat, n, i, ilast, ipos, ww;
     u_int32_t wd1,wd2;
     //numb32=numb/4;
     numb32=numb;
     ipos=0;
     ilast=32/adbit;
     pat=255;
     if (adbit == 1) {pat=1;}
     if (adbit == 2) {pat=3;}
     if (adbit == 4) {pat=15;}
     // if (adbit == 8) {pat=255;} 

     //for( n=0; n < numb32-1; n++)
     for( n=0; n < numb32; n++)
     {
        wd1=ia[n];
        wd2=ia[n+1];
        /* check sync pattern */
        //if ( (wd1 == 0xFFFFFFFFUL) && ((wd2 & 0xFF000000UL) == 0xFF000000UL) )  // sync detected
//        if ( (wd1 == 0xFFFFFFFFUL) && ((wd2 & 0xFF000000UL) == 0x8B000000UL) )  // sync detected 02-07-17
//         {      //printf("Sync detected ");   // removed  on 2002-10-3 
//                goto exit1;
//         }
        for(i=0; i<ilast; i++)
        {
             ww=(int)(wd1 & pat);
             xdata[ipos]=ww;
             wd1=wd1 >> adbit ; 
             ipos++;
         }
      // exit1:  ;
     }
     *npos=ipos;
}


void ymdhms2vdiftime(timeval_vdif_t *t)
{
/*  
      convert normal time to VDIF time
	       T.Kondo   2014-05-18
		   
*/
	//int year0, month0;
	int day0=1;
	int mjd0;
	//double d0;
	int days, secs;
    timeval_vdif_t twork;

	//get ref epoch

	t->ref_epo=(t->year-2000)*2;
	if(t->month >=6) t->ref_epo++;

	// days from ref epoch 
	twork=*t;
	twork.mday=1;
	if(twork.month >=6){
		twork.month=6;
	} else {
		twork.month=1;
	}
    ut2mjd_v(&twork);
	mjd0=twork.mjd;

	twork.month=t->month;
	twork.mday=t->mday;
    ut2mjd_v(&twork);

	days=twork.mjd-mjd0;

	// seconds from epoch

	secs=days*86400+t->hh*3600+t->mm*60+t->ss;

	t->sec_fm_ref=secs;

}

void vdif_header_make_ini(u_int32_t *head, vdif_header_t *vhead)
{
/*
      VDIF header generation (VDIF Kashima format)
	  initial call
	  2014-05-18
*/

	head[0]=0x0;
	head[0]=(head[0] & 0x7FFFFFFF) | ((vhead->invalid & 0x1) << 31);
	head[0]=(head[0] & 0xBFFFFFFF) | ((vhead->legacy & 0x19) << 30);
	head[0]=(head[0] & 0xC0000000) | (vhead->sec_fm_ref & 0x3FFFFFFF);

	head[1]=0x0;
	head[1]=(head[1] & 0xC0FFFFFF) | ((vhead->ref_epo & 0x3F) << 24);
	head[1]=(head[1] & 0xFF000000) | (vhead->frame & 0xFFFFFF);

	head[2]=0x0;
	head[2]=(head[2] & 0x1FFFFFFF) | ((vhead->ver & 0x7) << 29);
	head[2]=(head[2] & 0xE0FFFFFF) | ((vhead->log2_chs & 0x1F) << 24);
	head[2]=(head[2] & 0xFF000000) | (vhead->frame_len64 & 0xFFFFFF);

	head[3]=0x0;
	head[3]=(head[3] & 0x7FFFFFFF) | ((vhead->d_type & 0x1) << 31);
	head[3]=(head[3] & 0x83FFFFFF) | ((vhead->bpsampl & 0x1F) << 26);
	head[3]=(head[3] & 0xFC00FFFF) | ((vhead->thread_id & 0x3FF) << 16);
	head[3]=(head[3] & 0xFFFF0000) | (vhead->station_id & 0xFF);

	head[4]=0x0;
	head[4]=(head[4] & 0x00FFFFFF) | ((vhead->edv & 0xFF) << 24);

	// VDIF KASHIMA
	head[4]=(head[4] & 0xFF7FFFFF) | ((vhead->uflag & 0x1F) << 23);
	head[4]=(head[4] & 0xFF800000) | (vhead->srate & 0x7FFFFF);

	head[5]= vhead->sync_block;
	strncpy((char *)head+24,vhead->das,8);

}

void vdif_header_make(u_int32_t *head, vdif_header_t *vhead)
{
/*
      VDIF header generation (VDIF Kashima format)
	  only possible updated part
	  2014-05-18
*/

	head[0]=(head[0] & 0x7FFFFFFF) | ((vhead->invalid & 0x1) << 31);
	head[0]=(head[0] & 0xC0000000) | (vhead->sec_fm_ref & 0x3FFFFFFF);

	head[1]=(head[1] & 0xC0FFFFFF) | ((vhead->ref_epo & 0x3F) << 24);
	head[1]=(head[1] & 0xFF000000) | (vhead->frame & 0xFFFFFF);

	//head[2]=(head[2] & 0x1FFFFFFF) | ((vhead->ver & 0x7) << 29);
	//head[2]=(head[2] & 0xE0FFFFFF) | ((vhead->log2_chs & 0x1F) << 24);
	//head[2]=(head[2] & 0xFF000000) | (vhead->frame_len64 & 0xFFFFFF);

	//head[3]=(head[3] & 0x7FFFFFFF) | ((vhead->d_type & 0x1) << 31);
	//head[3]=(head[3] & 0x83FFFFFF) | ((vhead->bpsampl & 0x1F) << 26);
	head[3]=(head[3] & 0xFC00FFFF) | ((vhead->thread_id & 0x3FF) << 16);
	//head[3]=(head[3] & 0xFFFF0000) | (vhead->station_id & 0xFF);

	//head[4]=(head[4] & 0x00FFFFFF) | ((vhead->edv & 0xFF) << 24);

	// VDIF KASHIMA
	//head[4]=(head[4] & 0xFF7FFFFF) | ((vhead->uflag & 0x1F) << 23);
	//head[4]=(head[4] & 0xFF800000) | (vhead->srate & 0x0x7FFFFF);

	//head[5]= vhead->sync_block;
	//strncpy((char *)head+24,vhead->das,8);

}


void setup_outbuf_length(k5tovdifrunp_t *runp)
{
/*

    set obuf length

   [in]
        nframe_bytes  --- given frame length in byte
		nframe_sec    --- given # of frames per second  
		                  either nframe_bytes or nframe_sec is set
	    fsampl       ---- given sampling frequency
		adbit        ---- given AD bit
		totch        ---- given total channles (1-16)

   [return]
        nbuf         --- VDIF frame length including header (in 32bit-word)
        frame_bytes  --- VDIF frame length without header (bytes)
        frame_sec    --- # of frames in second (=imax)
        k5nbuf       --- Reading buffer size for K5 (in 32bt-word)
        unitp        --- Perod of frame length (sec)
        imax         --- # of frames in second
        usampl       --- # of samples/ch in a frame
    
*/
    int wspersec,wsperframe,wmax,wres,wsuggested;
    int wk5bytes;
	double fsample;

	/* 1st step set VDIF buffer size */

	//fsample=(double)runp->fmhz*1.0e6;
	fsample=(double)runp->fsampl;
	wspersec=(int)fsample*runp->adbit*runp->totch;

	if(runp->nframe_sec==0){ // check frame length
	   if(runp->frame_bytes+runp->headlen*4 > FRAME_LENGTH_MAX){
			printf(" Given frame length is over the maximum (%d) defined in the program!\n", FRAME_LENGTH_MAX);
			exit(0);
	   }
	   wsperframe=runp->frame_bytes*8;
	   wmax=wspersec/wsperframe;
	   wres=wspersec % wsperframe;
	   if(wres !=0){
		   printf(" # of samples per second (%d) devided by # of samples per frame (%d)\n",wspersec,wsperframe);
		   printf(" is not an integer!!\n");
		   wsuggested=wsperframe;
		   while((wspersec % wsuggested) !=0 ) {
			   runp->frame_bytes--;
			   wsuggested=runp->frame_bytes*8;
		   }
		   printf(" Advised frame length (without header) is %d bytes\n", runp->frame_bytes);
		   printf(" Try again with changing frame length parameter!\n");
		   exit(0);
	   }
	   runp->imax=wmax;
	   runp->nframe_sec=runp->imax;
	} else {  // calc # bytes per frame 
	   wres=(wspersec/32) % runp->nframe_sec;
	   if(wres !=0){
		   printf(" # of data (in 32-bit words) per second (%d)\n", wspersec/32);
		   printf("             devided by # of frames per second (%d)\n",runp->nframe_sec);
		   printf(" is not an integer!!\n");
		   wsuggested=runp->nframe_sec;
		   while(((wspersec/32) % wsuggested) !=0 ) {
			  runp->nframe_sec--;
			  wsuggested=runp->nframe_sec;
		   }
		   printf(" Advised # of frames per second is %d bytes\n", runp->nframe_sec);
		   printf(" Try again with changing # of frames per second!\n");
		   exit(0);
	   }
	   runp->frame_bytes=((wspersec/32)/runp->nframe_sec)*4;
	   runp->imax=runp->nframe_sec;
	}
	/* output buffer length */
	runp->nbuf=runp->frame_bytes/4 + runp->headlen;

	/* 2nd step  K5 reading buffer size setup */
    wk5bytes=runp->frame_bytes/runp->k5num;
    // check condition
    wspersec=(int)fsample*runp->adbit*runp->numch;
	wsperframe=wk5bytes*8;
    if( (wspersec % wsperframe) != 0){
        printf(" For K5 data reading,\n");
        printf(" # of samples per second (%d) devided by # of samples per frame (%d)\n",wspersec,wsperframe);
		printf(" is not an integer!!\n");
        exit(0);
    }
    runp->k5nbuf=wk5bytes/4;
    runp->unitp=1.0/(double)runp->imax;
    runp->usampl=(runp->frame_bytes*8)/(runp->adbit*runp->totch);

}

void combine_dir_name(char *fullpath, char *wdir, char *wname)
{
/*
     combine directory and name to get full path.
*/

	int i;
    strcpy(fullpath,wdir);
    i=strlen(fullpath);
    if(fullpath[i-1]!='\\' && fullpath[i-1]!='/'){
#ifdef WIN32
	   fullpath[i]='\\';
#else
	   fullpath[i]='/';
#endif
       fullpath[i+1]='\0';
	 }
	 strcat(fullpath,wname);
}

void setup_vdif_name(k5tovdifrunp_t *runp)
{
	   /* set up VDIF file name */

	char wname[100],wdir[256],odir[256];

    parse_dir_and_name(runp->k5name[0],odir,wname);
	if(runp->vdifname[0] == '\0'){  // VDIF name is K5 group 'a' name (extension is replaced by "vdif"
	   strcpy(runp->vdifname,runp->k5name[0]);
	   replace_extension(wname, "vdif");
	   if(runp->outdir[0] != '\0'){ // use given out directory
		   combine_dir_name(runp->vdifname, runp->outdir, wname);
	   } else {                    // use K5 directory
		   combine_dir_name(runp->vdifname, odir, wname);
	   }
	} else {                            // VDIF name is given
       parse_dir_and_name(runp->vdifname,wdir,wname);
	   if( runp->outdir[0]!='\0'){               // use given out directory
	      combine_dir_name(runp->vdifname, runp->outdir, wname);
	   } else {
	      if(wdir[0]=='\0'){  // use K5 file directory
	         combine_dir_name(runp->vdifname, odir, wname);
		  }
	   }
	}

}


int get_K5_info(k5tovdifrunp_t *runp, K5VSSP_file_t *k5)
{
       /* K5 data check to get sampling information */

	int i,n;
	double dwork;
    vsspaux_t aux;

    runp->totch=0;
    runp->bitsample=0;
    for(n=0; n<runp->k5num; n++){
       /* K5 type variable initialize */ 
       k5vssp_t_ini(&k5[n]);
	   k5[n].flag_fileclose=0;  // 2008-07-04
       /* Data file open */
       //k5[n].ibuf  = (unsigned char *)malloc(numb);             
       //k5[n].fname=runp.k5name[n];
       strcpy(k5[n].fname,runp->k5name[n]);
       if( (k5[n].stream  = fopen( runp->k5name[n], "rb" )) == NULL ){
           printf("Data File %s   Open Error! \n",runp->k5name[n]);
           return -1;
       }
       k5[n].size=get_file_size(k5[n].fname);

       /************************************************/
       /* Data header check to get sampling parameters */
       /************************************************/

       if(CheckK5header_with_Read(&k5[n]) < 0){
          return -1;
       }

       for(i=0; i<k5[n].numch; i++){  // set up K5 file ID and channel ID for K5 channel# (1-16) 
           runp->K5file_id[runp->totch+i]=n;       // 0-3
           runp->K5chan_id[runp->totch+i]=i;       // 0-3
       }

       runp->totch+=k5[n].numch;
       runp->bitsample+=k5[n].numch*k5[n].adbit;
       //printf("%d VSSP format %d\n",k5[n].seconds,k5[n].format);  // debug

       dwork=1000.0*(double)k5[n].sfreq*(double)(k5[n].adbit*k5[n].numch)/8.0; // bytes/sec without header
       dwork+=(double)k5[n].headsizew*4.0;  // bytes/sec including header
       runp->sec_period[n]=(double)k5[n].size/dwork;   // data length in second frame unit
       runp->isec_period[n]=(int)runp->sec_period[n];

       printf("Second Length :  %f\n",runp->sec_period[n]); // debug
    }

    if(k5[0].format != 2) {  // K5/VSSP
        if(runp->yyyy==0){
            printf("  This is K5/VSSP file. So please set DATE by -date option!\n");
            return -1;
        }
    }
    // AUX field parameter for the 1st K5 file

    if(k5[0].format==2){   // K5/VSSP32

//           printf("\n This is K5/VSSP32 format data.\n");
//           printf("\n  1st Extended Header Info is as follows.\n");
//           printf("    Year = %d    Total Day = %d\n", k5[0].year,k5[0].totalday);
//           printf("    Version  Major = %d  Minor = %d    # of AUX Bytes = %d\n",
//               k5[0].version1,k5[0].version2,k5[0].auxbyte);
//           printf("    AUX Field Data : ");
//           for(n=0; n<k5[0].auxbyte/4; n++){
//               printf(" %04X", k5[0].aux32[n]);
//           }

          runp->yyyy=k5[0].year;
          runp->ddd=k5[0].totalday;
          //get month day
          ddd2md(runp->yyyy,runp->ddd,&runp->month,&runp->dd);
          //get MJD
          runp->mjd=ut2mjd(&dwork,runp->yyyy,runp->month,runp->dd,0,0,0);

          runp->mjd3 = runp->mjd % 1000;



           aux.size=k5[0].auxbyte;
           //strncpy(aux.bdat,(char *)K5.aux32,K5.auxbyte); // this is not good for in case of data 0 is included
           for(n=0; n<aux.size; n++) aux.bdat[n]=*((char *)k5[0].aux32+n);
           vssp_aux_parse(&aux);
//           printf("\n        Format # = %d\n",aux.fmt);
//           if(aux.fmt != 0){
//             printf("        Filter (MHz) : %d\n",aux.filter);
//           }
           if(aux.fmt == 1){ // set station ID and name
               if(runp->sid[0]=='\0'){
                  runp->sid[0]=aux.sid2[0]; runp->sid[1]=aux.sid2[1];   // station ID
               }
               if(runp->das[0]=='\0'){
                   strncpy(runp->das,aux.station8,strlen(aux.station8));
               }
               //printf("        Station info : ID2(%s) name(%s)\n",aux.sid2,aux.station8);
           }
           if(aux.fmt == 2 ){   // set station name by PC host name
               if(runp->das[0]=='\0'){
                   strncpy(runp->das,aux.hostname8,strlen(aux.hostname8));
               }
//             printf("        PC host name :  %s\n",aux.hostname8);
           }

       }
//       printf("\n");



	return 0;
}


int setup_k5_file_position(k5tovdifrunp_t *runp, K5VSSP_file_t *k5)
{
/*
     if start time was different each file, then atamadashi is necessary
*/
    int n,aflag;
    int isa,secmin,secmax,common_sec_length;

    /* check start time */
    aflag=0;   // atamadashi flag
    secmin=k5[0].seconds;
    secmax=k5[0].seconds;
    if(runp->k5num > 0) {
        for(n=1; n<runp->k5num; n++){
            if(k5[n].seconds < secmin) secmin=k5[n].seconds;
            if(k5[n].seconds > secmax) secmax=k5[n].seconds;
        }
        if(secmin != secmax){
            aflag=1;
            printf(" Warning!! start time different file exists!\n");
        }
     }

     if(aflag==1){ // atamadashi at secmax
         for(n=0; n<runp->k5num; n++){
              isa= secmax-k5[n].seconds;
              runp->isec_period[n]-=isa;
              printf("atamadashi %d\n",isa);
              if(fwdNsecK5(isa, &k5[n]) <0){
                 return -1;
              }
          }
     }

     // find common data length in sec 
     common_sec_length=runp->isec_period[0];
     for(n=0; n<runp->k5num; n++){
         if(common_sec_length > runp->isec_period[n]) common_sec_length=runp->isec_period[n];
     }


     // start offset
     if(runp->soffset > 0){
         for(n=0; n<runp->k5num; n++){
            isa= runp->soffset;
            printf("File %d forward %d sec    ",n+1,isa);
            if(fwdNsecK5(isa, &k5[n]) <0){
               return -1;
            }
         }
         common_sec_length-=runp->soffset;
         printf("\n");
     }


     // set actual data conversion period if set so
     if(runp->adjust_flag==1){
         if(runp->period > 0 ) {
             if( runp->period > common_sec_length) runp->period=common_sec_length;
         } else {
             runp->period=common_sec_length;
         }
         if(runp->period == 0){
              printf(" K5 data length is less than 1 sec. So conversion to Mark-5B is cancelled!\n");
              return -1;
         }
     } 
     return 0;

}


void setup_vhead(k5tovdifrunp_t *runp, vdif_header_t *vhead)
{
/*
       set up vhead
*/

	   vhead->frame_len64=runp->nbuf/2; 

	   // set VDIF time

	   vhead->t.year=runp->yyyy;
	   vhead->t.month=runp->month;
	   vhead->t.mday=runp->dd;
	   vhead->t.hh=runp->hh;
	   vhead->t.mm=runp->mm;
	   vhead->t.ss=runp->ss;

	   // convert to VDIF time
	   ymdhms2vdiftime(&vhead->t);

	   // VDIF header fixed parameters
	   vhead->invalid=0;
	   vhead->legacy=0;
	   vhead->frame=0;
	   vhead->ver=0;
       vhead->log2_chs=(int)(log10((double)runp->totch)/log10(2.0));    
	   vhead->d_type=0;
       vhead->bpsampl=runp->adbit-1;  // means 1 bit sampling
	   vhead->station_id=0;
	   //strcpy((char *)vhead.station_id,"kk");
	   vhead->edv=EDV_NUMBER;    // means VDIF Kashima Format
	   vhead->uflag=1;  // meand MHz
	   vhead->srate=(int)runp->fmhz;
	   vhead->sync_block=VDIF_KASHIMA_SYNC;
	   strcpy(vhead->das,runp->das);

}


int main(int argc, char *argv[])
{
       int i,k,n,m;
       int ch;
       int iread,iwrite;
       int kpos;
       unsigned int iw;
       //int obufmax=225;
       int ow;

       u_int32_t *obuf;  // for 1 Frame data 
       u_int32_t *head;   // for header 8 X 32 bits
       u_int32_t *idata;  // for data field 
       
       int nframe;
       int nsample32;   // # of samples per 32 bits
       int period;
       int tmax,tmin;
       int sec_count; 
       int f_id,ch_id;
       int flag_fo_open; 
       double telapsed;           // to measure processing time
       time_t tstart, tfinish;    
       FILE *fo;
       k5tovdifrunp_t runp;
       K5VSSP_file_t k5[4];   // IP-VLBI data type variables 
       vdif_header_t vhead;
       int sync_pat=VDIF_KASHIMA_SYNC;  // VDIF Kashima format sync pattern 0xACABFEED
       int imask[]={0,0x7FFFFFFF,0x3FFFFFFF,0,0xFFFFFFF,0,0,0,0xFFFFFF};
                 // masking table for 1,2,4,8 bit sampling data
       int *conv2bit;
       int conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)
       int conv2bit_2[]={0,2,1,3};   //  conversion 2 bit (default)
       int conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       } 
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for VDIF data\n",VDIFDIR);
          printf("                      ( %s )\n",getenv(VDIFDIR));
          printf("\n");
          exit(0);
          }
       }

       strcpy(runp.program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_k5tovdif(argc, argv, &runp); 

       if(runp.e2bit==1){
           conv2bit=conv2bit_1;
       } else if(runp.e2bit==3){
           conv2bit=conv2bit_3;
       } else {
           conv2bit=conv2bit_2;
       }


	   /* set up K5 full path name */

       setup_k5files_full_path(&runp);

       /* K5 data check to get sampling information */
       if(get_K5_info(&runp, k5)<0) goto exitend;



	   /* set up VDIF file name */
       setup_vdif_name(&runp);
       printf(" VDIF name : %s\n",runp.vdifname);

       runp.adjust_flag=1;   // set period is integer second mode  2010-07-21
       flag_fo_open=0;

       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    K5/VSSP to VDIF Data Format Converter                 *\n");
       printf("*      %s  by  T.KONDO/NICT    *\n",runp.program_version);
       if(runp.nocheck==1)
       printf("*               (!! No header check mode!!)                *\n");
       printf("************************************************************\n");
       printf("\n");

       //if(runp.monit==1)
//       monit_mk5info(&mk5info);  // debug    

//       if(runp.monit==1) monit_runp(&runp,&vhead);

       tstart= time(NULL);





        /********************************/
        /* set internal-use parameters  */
        /********************************/

        runp.sfreq=k5[0].sfreq;       // sampling frequency (kHz)
        runp.fmhz=runp.sfreq/1000.0;  
        runp.adbit=k5[0].adbit;       // A/D bits
        runp.numch=k5[0].numch;       // # of channel in data (1 or 4)
        runp.fsampl=1000.0*(double)runp.sfreq;   // sampling frequency (Hz)
        runp.tsampl=1.0/runp.fsampl;            // sampling period (sec)
        runp.bits=((int)runp.fsampl)*runp.adbit*runp.numch;  // # of bits in 1 sec data except header
        runp.bytes=runp.bits/8;                     // # of bytes in 1 sec data except header

        runp.samplpp=(int)runp.fsampl;  // # of samples per ch in a second


//        if(runp.adbit > 2){
//           printf("AD bits (%d) exceeds 2 bits!\n",runp.adbit); 
//           printf("Not supported now!\n"); 
//           goto exitend;
//        }

        if( (32 % runp.bitsample) !=0) {
           printf(" # of bits per samples is %d\n",runp.bitsample); 
           printf(" So that # of samples in a 32-bit word is not integer!\n"); 
           printf(" This situation is not allowded in the current version.\n");
           goto exitend;
        }
        nsample32=32/runp.bitsample;  // number of samples per ch in a 32-bit word 

        //debug
        //printf("samplpp,totch, bitsample,nsample32  %d %d %d %d\n",
        //  samplpp, totch,bitsample, nsample32);


        /* set buffer length for K5 and VDIF file */

        setup_outbuf_length(&runp);

        /* set up K5 reading condition */

        // allocate reading buffer 
        for(n=0; n<runp.k5num; n++){
            k5[n].ubuf  = (unsigned int *)malloc(runp.k5nbuf*sizeof(int));
            k5[n].idata = (int *)malloc(runp.usampl*runp.numch*sizeof(int));
            k5[n].bytesec=runp.bytes;  
        }


       /* if start time was different each file, then atamadashi is necessary */
       if(setup_k5_file_position(&runp, k5)<0) goto exitend;

 
	   /* output buffer set */
       obuf  = (u_int32_t *)malloc(runp.nbuf*sizeof(u_int32_t)); 

       head=obuf;
       idata=obuf+8;

       /* setup vdif common header parameters */
       // reflect current K5 time
       sec2hms(k5[0].seconds, &runp.hh, &runp.mm, &runp.ss);
       setup_vhead(&runp, &vhead);


	   // VDIF header variable parameters
	   vhead.thread_id=0;
	   vhead.ref_epo=vhead.t.ref_epo;
	   vhead.sec_fm_ref=vhead.t.sec_fm_ref;

       /* run condition monitor */

       monit_runp(&runp, &vhead);

       /* VDIF file create */

       if(runp.vdifname[0]=='\0'){
           printf(" VDIF file not created!\n");
       } else {
          if( (fo  = fopen( runp.vdifname, "wb" )) == NULL ){
              printf("  File  %s  Open Error! \n",runp.vdifname);
              goto exitend;
          }
       }
       flag_fo_open=1;

       

	   // initialize VDIF header
       vdif_header_make_ini(head, &vhead);

	   //printf("Debug %08X %08X %08X %08X\n",head[0],head[1],head[2],head[3]);
	   //printf("      %08X %08X %08X %08X\n",head[4],head[5],head[6],head[7]);

       /* conversion start message */

       printf("\n");
       printf("K5 file(s)\n");
       k=1;  // K5 ch# start
       for(n=0; n<runp.k5num; n++){
           if(k5[n].numch==1){ // 1ch mode data
               printf(" FILE%d CH# %2d    : %s\n",n+1,k, k5[n].fname);
               k++;
           } else {  //4ch mode
               printf(" FILE%d CH# %2d-%2d : %s\n",n+1,k,k+3, k5[n].fname);
               k+=4;
           }
           printf("   (A/D(bits) %d  CHs %d  SFreq(kHz) %d",
                     k5[n].adbit,k5[n].numch,k5[n].sfreq);
           printf(" Time %02d:%02d:%02d  sec %d)",
                      k5[n].hh,k5[n].mm,k5[n].ss,k5[n].seconds);
           if(k5[n].format==1){                 // add 2007-04-24
               printf(" VSSP\n");
           } else if(k5[n].format==2){
               printf(" VSSP32\n");
           } else {
               printf("\n");
           }
       }

       printf("\n");

       printf("VDIF file (created) : %s\n",runp.vdifname);
       printf("    Observation Date : 3digit MJD = %4d\n",runp.mjd3);
       if(runp.yyyy != 0){
       printf("                     :      Year  = %4d\n",runp.yyyy);
       printf("                     :      Month = %4d\n",runp.mm);
       printf("                     :      Day   = %4d\n",runp.dd);
       printf("                     :  Total Day = %4d\n",runp.ddd);
       }
       if(runp.adjust_flag ==1){
       printf("    Conversion length (sec) :  %d\n",runp.period);
       }
       printf("\n");

       printf("                   TIME LABEL MONITOR\n");
       printf(" ==========================================================\n");   
       printf("  K5-FILE1  K5-FILE2  K5-FILE3  K5-FILE4      VDIF-TIME\n");   
       printf("  HH:MM:SS  HH:MM:SS  HH:MM:SS  HH:MM:SS   RR SSSSSSSSSS\n");   
       printf(" ----------------------------------------------------------\n");   


       period=0;
	   sec_count=0;


       ow=8;

       while(1){    // entry point of second loop

           nframe=0;  // frame # (in a second) counter reset
		   vhead.frame=0;
		   vhead.t.sec_fm_ref=vhead.sec_fm_ref;
           vdiftime2ymdhms(&vhead.t);   // get YYYY/MM/DD HH:MM:SS for debug
	       //printf("Debug %08X %08X %08X %08X\n",head[0],head[1],head[2],head[3]);  //debug
	       //printf("      %08X %08X %08X %08X\n",head[4],head[5],head[6],head[7]);  //debug
		   //print time label
//		   printf("  %04d/%02d/%02d %02d:%02d:%02d",vhead.t.year,vhead.t.month,vhead.t.mday,
//			   vhead.t.hh,vhead.t.mm,vhead.t.ss);
//		   printf("               %4d    %10d\n",vhead.ref_epo,vhead.sec_fm_ref);


           for(i=0; i<runp.imax; i++){  // 1 sec loop is imax'th VDIF frame

               vdif_header_make(head,&vhead); 
               /* 2012-12-03 new algorythm applied and moved from below */

               if(i==0){ //  print time label  2012-12-03
                   tmax=k5[0].seconds;
                   tmin=tmax;
                   for(n=0; n<runp.k5num; n++){
                       printf("  %02d:%02d:%02d",k5[n].hh,k5[n].mm,k5[n].ss);
                       if(k5[n].seconds > tmax) tmax=k5[n].seconds;
                       if(k5[n].seconds < tmin) tmin=k5[n].seconds;
                   }
                   if(runp.k5num != 4){
                      for(n=runp.k5num; n<4; n++) printf("          ");
                   }

                   printf("   %02d %010d",vhead.ref_epo,vhead.sec_fm_ref);
                   printf("\n");
			   }

               /* read K5 data corresponding to one frame of VDIF */ 
               for(n=0; n<runp.k5num; n++){
                  //iread = fread( k5[n].ibuf, sizeof( char ), numb, k5[n].stream);
                   iread = fread( k5[n].ubuf, sizeof(int), runp.k5nbuf, k5[n].stream);
                   if (iread <= 0) goto exit_loop;  // exit loop  2008-07-04
                       // all K5 file must be the same sampling condition
                   parseADdataKOKO(k5[n].ubuf,runp.k5nbuf,runp.adbit,k5[n].idata, &kpos);
                  //parseADdata_3N(&k5[n]);
               }
               //kpos=k5[0].kpos;

               //printf("Debug numb kpos %d %d\n",numb,kpos);

               /* convert to VDIF frame data */


               // frame data generation
               ow=8;
               for(k=0; k<runp.frame_bytes/4; k++){
                   for(m=0; m<nsample32; m++){
                       for(ch=0; ch<runp.totch; ch++){
                           f_id=runp.K5file_id[runp.ch_assign[ch]-1];
                           ch_id=runp.K5chan_id[runp.ch_assign[ch]-1];
                           iw=k5[f_id].idata[ch_id+k5[f_id].numch*(k*nsample32+m)];
                           if(runp.adbit==2) iw=conv2bit[iw]; // code conversion in case of 2 bit
                           iw <<= (32-runp.adbit);
                           idata[k]=((idata[k] >> runp.adbit) & imask[runp.adbit]) | iw;
                       }
                   }
                   ow++;
               }
               iwrite=fwrite(obuf,sizeof(int),runp.nbuf,fo);


               /* frame # update */

               nframe++;
			   vhead.frame++;

           }
           sec_count++; 
		   vhead.sec_fm_ref++;

           // second data read end

           period++;
           if(runp.period > 0 && (runp.period == period)) goto exit_loop;

           // K5 header read
           tmax=k5[0].seconds;
           tmin=tmax;
           for(n=0; n<runp.k5num; n++){
              if(runp.nocheck==1){  // 2010-08-03
                 if(CheckK5header_with_Read_but_no_search(&k5[n])<0){
                     k5[n].flag_fileclose=1;
                     goto exit_loop;  // already file is closed
                 }
              } else {
                 if(CheckK5header_with_Read(&k5[n]) < 0){
                     k5[n].flag_fileclose=1;
                     goto exit_loop;  // already file is closed
                 }
              }
           } 
       }

exit_loop:

       printf(" ==========================================================\n");   

       for(n=0; n<runp.k5num; n++){
		   if(k5[n].flag_fileclose==0) i=fclose(k5[n].stream);
       }

       if(ow < runp.nbuf){
            iwrite=fwrite(obuf,sizeof(int),ow,fo);
       }

       i=fclose(fo);


       printf("\n");
       printf("  VDIF file (%s) has been created.\n\n",runp.vdifname);



       free(obuf);

       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
       return 0;
}
